package handler

import (
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"io/ioutil"
	"strings"
	"time"

	"github.com/checkr/flagr/pkg/config"
	"github.com/checkr/flagr/pkg/util"
	"github.com/checkr/flagr/swagger_gen/models"

	"github.com/Shopify/sarama"
	"github.com/sirupsen/logrus"
)

var (
	saramaNewAsyncProducer = sarama.NewAsyncProducer
)

func mustParseKafkaVersion(version string) sarama.KafkaVersion {
	v, err := sarama.ParseKafkaVersion(version)
	if err != nil {
		panic(err)
	}
	return v
}

// NewKafkaRecorder creates a new Kafka recorder
var NewKafkaRecorder = func() DataRecorder {
	cfg := sarama.NewConfig()
	tlscfg := createTLSConfiguration(
		config.Config.RecorderKafkaCertFile,
		config.Config.RecorderKafkaKeyFile,
		config.Config.RecorderKafkaCAFile,
		config.Config.RecorderKafkaVerifySSL,
	)
	if tlscfg != nil {
		cfg.Net.TLS.Enable = true
		cfg.Net.TLS.Config = tlscfg
	}
	cfg.Producer.RequiredAcks = sarama.WaitForLocal
	cfg.Producer.Retry.Max = config.Config.RecorderKafkaRetryMax
	cfg.Producer.Flush.Frequency = config.Config.RecorderKafkaFlushFrequency
	cfg.Version = mustParseKafkaVersion(config.Config.RecorderKafkaVersion)

	brokerList := strings.Split(config.Config.RecorderKafkaBrokers, ",")
	producer, err := saramaNewAsyncProducer(brokerList, cfg)
	if err != nil {
		logrus.WithField("kafka_error", err).Fatal("Failed to start Sarama producer:")
	}

	// We will just log to STDOUT if we're not able to produce messages.
	if producer != nil {
		go func() {
			for err := range producer.Errors() {
				logrus.WithField("kafka_error", err).Error("failed to write access log entry")
			}
		}()
	}

	var encryptor dataRecordEncryptor
	if config.Config.RecorderKafkaEncrypted && config.Config.RecorderKafkaEncryptionKey != "" {
		encryptor = newSimpleboxEncryptor(config.Config.RecorderKafkaEncryptionKey)
	}

	return &kafkaRecorder{
		topic:    config.Config.RecorderKafkaTopic,
		producer: producer,
		options: DataRecordFrameOptions{
			Encrypted:       config.Config.RecorderKafkaEncrypted,
			Encryptor:       encryptor,
			FrameOutputMode: config.Config.RecorderFrameOutputMode,
		},
	}
}

func createTLSConfiguration(certFile string, keyFile string, caFile string, verifySSL bool) (t *tls.Config) {
	if certFile != "" && keyFile != "" && caFile != "" {
		cert, err := tls.LoadX509KeyPair(certFile, keyFile)
		if err != nil {
			logrus.WithField("TLSConfigurationError", err).Panic(err)
		}

		caCert, err := ioutil.ReadFile(caFile)
		if err != nil {
			logrus.WithField("TLSConfigurationError", err).Panic(err)
		}

		caCertPool := x509.NewCertPool()
		caCertPool.AppendCertsFromPEM(caCert)

		t = &tls.Config{
			Certificates:       []tls.Certificate{cert},
			RootCAs:            caCertPool,
			InsecureSkipVerify: !verifySSL,
		}
	}
	// will be nil by default if nothing is provided
	return t
}

type kafkaRecorder struct {
	producer sarama.AsyncProducer
	topic    string
	options  DataRecordFrameOptions
}

func (k *kafkaRecorder) NewDataRecordFrame(r models.EvalResult) DataRecordFrame {
	return DataRecordFrame{
		evalResult: r,
		options:    k.options,
	}
}

func (k *kafkaRecorder) AsyncRecord(r models.EvalResult) {
	frame := k.NewDataRecordFrame(r)
	output, err := frame.Output()
	if err != nil {
		logrus.WithField("err", err).Error("failed to generate data record frame for kafka recorder")
		return
	}
	k.producer.Input() <- &sarama.ProducerMessage{
		Topic:     k.topic,
		Key:       sarama.StringEncoder(frame.GetPartitionKey()),
		Value:     sarama.ByteEncoder(output),
		Timestamp: time.Now().UTC(),
	}

	logKafkaAsyncRecordToDatadog(r)
}

var logKafkaAsyncRecordToDatadog = func(r models.EvalResult) {
	if config.Global.StatsdClient == nil {
		return
	}
	config.Global.StatsdClient.Incr(
		"data_recorder.kafka",
		[]string{
			fmt.Sprintf("FlagID:%d", util.SafeUint(r.FlagID)),
		},
		float64(1),
	)
}
